import sys

import pytest

from aana.api.event_handlers.event_manager import EventManager
from aana.api.event_handlers.rate_limit_handler import RateLimitHandler
from aana.exceptions.runtime import TooManyRequestsException


def test_rate_limiter_single():
    """Tests that the rate limiter raises if the rate limit is exceeded."""
    event_manager = EventManager()
    rate_limiter = RateLimitHandler(1, 1.0)

    event_manager.register_handler_for_events(rate_limiter, ["foo"])
    event_manager.handle("foo")
    with pytest.raises(TooManyRequestsException):
        event_manager.handle("foo")


def test_rate_limiter_multiple():
    """Tests that the rate limiter raises if the rate limit is exceeded."""
    event_manager = EventManager()
    rate_limiter = RateLimitHandler(1, 1.0)

    event_manager.register_handler_for_events(rate_limiter, ["foo", "bar"])
    event_manager.handle("foo")
    with pytest.raises(TooManyRequestsException):
        event_manager.handle("bar")


def test_rate_limiter_noraise():
    """Tests that the rate limiter raises if the rate limit is exceeded."""
    event_manager = EventManager()
    # Smallest possible value such that 1+x>1, should never run into rate limit
    rate_limiter = RateLimitHandler(1, sys.float_info.epsilon)
    event_manager.register_handler_for_events(rate_limiter, ["foo"])
    for _ in range(10):
        event_manager.handle("foo")


def test_event_manager_discriminates():
    """Tests that the rate limiter raises if the rate limit is exceeded."""
    event_manager = EventManager()
    rate_limiter = RateLimitHandler(1, sys.float_info.max)
    event_manager.register_handler_for_events(rate_limiter, ["bar"])
    # Should not raise
    event_manager.handle("foo")



import collections.abc

from pydantic import BaseModel, ConfigDict, Field, field_validator


class MyConfig(ConfigDict, total=False):
    json_schema_extra: dict


__all__ = ["WhisperParams", "BatchedWhisperParams"]


class WhisperParams(BaseModel):
    """A model for the Whisper audio-to-text model parameters.

    Attributes:
        language (str): Optional language code such as "en" or "fr".
                        If None, language will be automatically detected.
        beam_size (int): Size of the beam for decoding.
        best_of (int): Number of best candidate sentences to consider.
        temperature (Union[float, List[float], Tuple[float, ...]]): Controls the sampling
            randomness.  It can be a tuple of temperatures,
            which will be successively used upon failures according to either
            [compression_ratio_threshold](https://github.com/guillaumekln/faster-whisper/blob/5a0541ea7d054aa3716ac492491de30158c20057/faster_whisper/transcribe.py#L216) or
            [log_prob_threshold](https://github.com/guillaumekln/faster-whisper/blob/5a0541ea7d054aa3716ac492491de30158c20057/faster_whisper/transcribe.py#L218C23-L218C23).
        word_timestamps (bool): Whether to extract word-level timestamps.
        vad_filter (bool): Whether to enable voice activity detection to filter non-speech.
    """

    language: str | None = Field(
        default=None,
        description="Language code such as 'en' or 'fr'. If None, language is auto-detected.",
    )
    beam_size: int = Field(
        default=5, ge=1, description="Size of the beam to use for decoding."
    )
    best_of: int = Field(
        default=5, ge=1, description="Number of best candidate sentences to consider."
    )
    temperature: float | collections.abc.Sequence[float] = Field(
        default=(0.0, 0.2, 0.4, 0.6, 0.8, 1.0),
        description=(
            "Temperature for sampling. A single value or a sequence indicating fallback temperatures."
        ),
    )
    without_timestamps: bool = Field(
        default=False,
        description="Whether to sample only text tokens in decoding.",
    )
    word_timestamps: bool = Field(
        default=False, description="Whether to extract word-level timestamps."
    )
    vad_filter: bool = Field(
        default=True,
        description="Whether to enable voice activity detection filtering.",
    )

    @field_validator("temperature")
    @classmethod
    def check_temperature(cls, v: float):
        """Validates a temperature value.

        Args:
            v (float): Value to validate.

        Raises:
            ValueError: Temperature is out of range.

        Returns:
            Temperature value.
        """
        if isinstance(v, float) and not 0 <= v <= 1:
            raise ValueError(  # noqa: TRY003
                "Temperature must be between 0 and 1 when a single float is provided."
            )
        if isinstance(v, list | tuple) and not all(0 <= t <= 1 for t in v):
            raise ValueError(  # noqa: TRY003
                "Each temperature in the sequence must be between 0 and 1."
            )
        return v

    model_config = MyConfig(
        json_schema_extra={
            "description": "Parameters for the Whisper audio-to-text model."
        }
    )


class BatchedWhisperParams(BaseModel):
    """A model for the Batched version of Whisper audio-to-text model parameters.

    Attributes:
        language (str): Optional language code such as "en" or "fr".
                        If None, language will be automatically detected.
        beam_size (int): Size of the beam for decoding.
        best_of (int): Number of best candidate sentences to consider.
        temperature (Union[float, List[float], Tuple[float, ...]]): Controls the sampling
            randomness.  It can be a tuple of temperatures,
            which will be successively used upon failures according to either
            [compression_ratio_threshold](https://github.com/guillaumekln/faster-whisper/blob/5a0541ea7d054aa3716ac492491de30158c20057/faster_whisper/transcribe.py#L216) or
            [log_prob_threshold](https://github.com/guillaumekln/faster-whisper/blob/5a0541ea7d054aa3716ac492491de30158c20057/faster_whisper/transcribe.py#L218C23-L218C23).
        #TODO: add other parameters
    """

    language: str | None = Field(
        default=None,
        description="Language code such as 'en' or 'fr'. If None, language is auto-detected.",
    )
    beam_size: int = Field(
        default=5, ge=1, description="Size of the beam to use for decoding."
    )
    best_of: int = Field(
        default=5, ge=1, description="Number of best candidate sentences to consider."
    )
    temperature: float | collections.abc.Sequence[float] = Field(
        default=(0.0, 0.2, 0.4, 0.6, 0.8, 1.0),
        description=(
            "Temperature for sampling. A single value or a sequence indicating fallback temperatures."
        ),
    )
    without_timestamps: bool = Field(
        default=True,
        description="Whether to sample only text tokens in decoding. Set to 'False' to get phrase-level time stamps (eg. for subtitles)",
    )
    word_timestamps: bool = Field(
        default=False, description="Whether to extract word-level timestamps."
    )

    @field_validator("temperature")
    @classmethod
    def check_temperature(cls, v: float):
        """Validates a temperature value.

        Args:
            v (float): Value to validate.

        Raises:
            ValueError: Temperature is out of range.

        Returns:
            Temperature value.
        """
        if isinstance(v, float) and not 0 <= v <= 1:
            raise ValueError(  # noqa: TRY003
                "Temperature must be between 0 and 1 when a single float is provided."
            )
        if isinstance(v, list | tuple) and not all(0 <= t <= 1 for t in v):
            raise ValueError(  # noqa: TRY003
                "Each temperature in the sequence must be between 0 and 1."
            )
        return v

    model_config = MyConfig(
        json_schema_extra={
            "description": "Parameters for the Batched Whisper audio-to-text model."
        }
    )



from typing import TypeVar

from sqlalchemy.orm import Session

from aana.core.models.captions import Caption, CaptionsList
from aana.storage.models.caption import CaptionEntity
from aana.storage.repository.base import BaseRepository

T = TypeVar("T", bound=CaptionEntity)


class CaptionRepository(BaseRepository[T]):
    """Repository for Captions."""

    def __init__(self, session: Session, model_class: type[T] = CaptionEntity):
        """Constructor."""
        super().__init__(session, model_class)

    def save(self, model_name: str, caption: Caption, timestamp: float, frame_id: int):
        """Save a caption.

        Args:
            model_name (str): The name of the model used to generate the caption.
            caption (Caption): The caption.
            timestamp (float): The timestamp.
            frame_id (int): The frame ID.
        """
        entity = CaptionEntity.from_caption_output(
            model_name=model_name,
            frame_id=frame_id,
            timestamp=timestamp,
            caption=caption,
        )
        self.create(entity)
        return entity

    def save_all(
        self,
        model_name: str,
        captions: CaptionsList,
        timestamps: list[float],
        frame_ids: list[int],
    ) -> list[CaptionEntity]:
        """Save captions.

        Args:
            model_name (str): The name of the model used to generate the captions.
            captions (CaptionsList): The captions.
            timestamps (list[float]): The timestamps.
            frame_ids (list[int]): The frame IDs.

        Returns:
            list[CaptionEntity]: The list of caption entities.
        """
        entities = [
            CaptionEntity.from_caption_output(
                model_name=model_name,
                frame_id=frame_id,
                timestamp=timestamp,
                caption=caption,
            )
            for caption, timestamp, frame_id in zip(
                captions, timestamps, frame_ids, strict=True
            )
        ]
        results = self.create_multiple(entities)
        return results



from typing import Any

from haystack import component
from pydantic import BaseModel
from ray import serve

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.utils.asyncio import run_async
from aana.utils.core import import_from_path


@component
class RemoteHaystackComponent:
    """A component that connects to a remote Haystack component created by HaystackComponentDeployment.

    Attributes:
        deployment_name (str): The name of the deployment to use.
    """

    def __init__(
        self,
        deployment_name: str,
    ):
        """Initialize the component.

        Args:
            deployment_name (str): The name of the deployment to use.
        """
        self.deployment_name = deployment_name

    def warm_up(self):
        """Warm up the component.

        This will properly initialize the component by creating a handle to the deployment
        and setting the input and output types.
        """
        if hasattr(self, "handle"):
            return
        self.handle = run_async(AanaDeploymentHandle.create(self.deployment_name))
        sockets = run_async(self.handle.get_sockets())
        component.set_input_types(
            self, **{socket.name: socket.type for socket in sockets["input"].values()}
        )
        component.set_output_types(
            self, **{socket.name: socket.type for socket in sockets["output"].values()}
        )

    def run(self, **data):
        """Run the component on the input data."""
        return run_async(self.handle.run(**data))


class HaystackComponentDeploymentConfig(BaseModel):
    """Configuration for the HaystackComponentDeployment.

    Attributes:
        component (str): The path to the Haystack component to deploy.
        params (dict): The parameters to pass to the component on initialization (model etc).
    """

    component: str
    params: dict[str, Any]


@serve.deployment
class HaystackComponentDeployment(BaseDeployment):
    """Deployment to deploy a Haystack component."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It creates the Haystack component and warms it up.

        The configuration should conform to the HaystackComponentDeploymentConfig schema.
        """
        config_obj = HaystackComponentDeploymentConfig(**config)

        self.params = config_obj.params
        self.component_path = config_obj.component
        self.component = import_from_path(config_obj.component)(**config_obj.params)

        self.component.warm_up()

    @exception_handler
    async def run(self, **data: dict[str, Any]) -> dict[str, Any]:
        """Run the model on the input data."""
        return self.component.run(**data)

    async def get_sockets(self):
        """Get the input and output sockets of the component."""
        return {
            "output": self.component.__haystack_output__._sockets_dict,
            "input": self.component.__haystack_input__._sockets_dict,
        }



from typing import TypeVar

from sqlalchemy.orm import Session

from aana.core.models.asr import (
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.storage.models.transcript import TranscriptEntity
from aana.storage.repository.base import BaseRepository

T = TypeVar("T", bound=TranscriptEntity)


class TranscriptRepository(BaseRepository[T]):
    """Repository for Transcripts."""

    def __init__(self, session: Session, model_class: type[T] = TranscriptEntity):
        """Constructor."""
        super().__init__(session, model_class)

    def save(
        self,
        model_name: str,
        transcription_info: AsrTranscriptionInfo,
        transcription: AsrTranscription,
        segments: AsrSegments,
    ) -> TranscriptEntity:
        """Save transcripts.

        Args:
            model_name (str): The name of the model used to generate the transcript.
            transcription_info (AsrTranscriptionInfo): The ASR transcription info.
            transcription (AsrTranscription): The ASR transcription.
            segments (AsrSegments): The ASR segments.

        Returns:
            TranscriptEntity: The transcript entity.
        """
        transcript_entity = TranscriptEntity.from_asr_output(
            model_name=model_name,
            transcription=transcription,
            segments=segments,
            info=transcription_info,
        )
        self.session.add(transcript_entity)
        self.session.commit()
        return transcript_entity



# ruff: noqa: S101, S113
import importlib
import uuid
from pathlib import Path

import pytest
import yaml
from click.testing import CliRunner

from aana.cli import cli

test_app_path = "aana.tests.projects.lowercase.app:aana_app"
port = 5000


@pytest.fixture
def config_paths():
    """Fixture to create config paths and clean them up after the test."""
    # Temporary file paths creation
    app_module, _ = test_app_path.split(":")
    output_dir = Path(importlib.util.find_spec(app_module).origin).parent

    app_config_name = f"app_config_{uuid.uuid4().hex}"
    config_name = f"config_{uuid.uuid4().hex}"

    app_config_path = output_dir / f"{app_config_name}.py"
    config_path = output_dir / f"{config_name}.yaml"

    # Yield paths to use in the test
    yield app_config_path, config_path

    # Cleanup code after test execution
    app_config_path.unlink(missing_ok=True)
    config_path.unlink(missing_ok=True)


# TODO: Fix the test
# class TestAppManager:
#     """Context manager to handle the test app process."""

#     def __init__(self, target):
#         """Initialize the context manager."""
#         self.process = multiprocessing.Process(target=target)

#     def __enter__(self):
#         """Start the process when entering the context manager."""
#         self.process.start()
#         return self.process

#     def __exit__(self, exc_type, exc_value, traceback):
#         """Terminate the process when exiting the context manager."""
#         self.process.terminate()


# def deploy_test_app():
#     """Starting the CLI application deployment."""
#     CliRunner().invoke(cli, ["deploy", test_app_path, "--port", str(port)])


# @pytest.mark.asyncio
# async def test_aana_deploy():
#     """Test aana deploy command."""
#     with TestAppManager(deploy_test_app):
#         timeout = 30  # timeout in seconds
#         url = f"http://localhost:{port}/api/ready"

#         # Polling the endpoint until the app is ready or timeout expires
#         for _ in range(timeout):
#             try:
#                 response = requests.get(url, timeout=1)
#                 if response.status_code == 200:
#                     break
#             except requests.exceptions.RequestException:
#                 pass

#             await asyncio.sleep(1)
#         else:
#             # Only raises if the for loop completes without breaking (i.e. timeout)
#             raise Exception("App not started within the given timeout")

#         assert response.status_code == 200
#         assert response.json() == {"ready": True}

#         # Test lowercase endpoint
#         data = {"text": ["Hello World!", "This is a test."]}
#         response = requests.post(
#             f"http://localhost:{port}/lowercase",
#             data={"body": json.dumps(data)},
#         )
#         assert response.status_code == 200
#         lowercase_text = response.json().get("text")
#         assert lowercase_text == ["hello world!", "this is a test."]


def test_aana_build(config_paths):
    """Test aana build command."""
    expected_app_config = (
        "from aana.tests.projects.lowercase.app import aana_app\n"
        "\n"
        "task_queue_deployment = aana_app.get_deployment_app('task_queue_deployment')\n"
        "lowercase_deployment = aana_app.get_deployment_app('lowercase_deployment')\n"
        "lowercase_app = aana_app.get_main_app()\n"
    )

    app_config_path, config_path = config_paths

    result = CliRunner().invoke(
        cli,
        [
            "build",
            test_app_path,
            "--port",
            str(port),
            "--app-config-name",
            app_config_path.stem,
            "--config-name",
            config_path.stem,
        ],
    )

    assert str(app_config_path) in result.output
    assert str(config_path) in result.output

    assert app_config_path.exists()
    assert config_path.exists()

    with app_config_path.open() as f:
        app_config = f.read()

    assert app_config == expected_app_config

    with config_path.open() as f:
        config = yaml.load(f, Loader=yaml.FullLoader)  # noqa: S506

    assert config["http_options"]["port"] == port
    assert "applications" in config
    assert len(config["applications"]) == 3
    for app in config["applications"]:
        assert app["name"] in [
            "lowercase_deployment",
            "lowercase_app",
            "task_queue_deployment",
        ]


def test_aana_missing_app_path():
    """Test that the CLI commands fail when the APP_PATH argument is missing."""
    result = CliRunner().invoke(cli, ["deploy"])
    assert result.exit_code == 2
    assert "Error: Missing argument 'APP_PATH'." in result.output

    result = CliRunner().invoke(cli, ["build"])
    assert result.exit_code == 2
    assert "Error: Missing argument 'APP_PATH'." in result.output

    result = CliRunner().invoke(cli, ["migrate"])
    assert result.exit_code == 2
    assert "Error: Missing argument 'APP_PATH'." in result.output



__all__ = ["get_gpu_memory"]


def get_gpu_memory(gpu: int = 0) -> int:
    """Get the total memory of a GPU in bytes.

    Args:
        gpu (int): the GPU index. Defaults to 0.

    Returns:
        int: the total memory of the GPU in bytes
    """
    import torch

    return torch.cuda.get_device_properties(gpu).total_memory



# ruff: noqa: S101

from typing import TypedDict

from aana.utils.typing import is_typed_dict


def test_is_typed_dict() -> None:
    """Test the is_typed_dict function."""

    class Foo(TypedDict):
        foo: str
        bar: int

    assert is_typed_dict(Foo) == True


def test_is_not_typed_dict():
    """Test the is_typed_dict function."""

    class Foo:
        pass

    assert is_typed_dict(Foo) == False
    assert is_typed_dict(dict) == False



# ruff: noqa: S101
import pytest

from aana.core.models.sampling import SamplingParams


def test_valid_sampling_params():
    """Test valid sampling parameters."""
    params = SamplingParams(
        temperature=0.5, top_p=0.9, top_k=10, max_tokens=50, repetition_penalty=1.5
    )
    assert params.temperature == 0.5
    assert params.top_p == 0.9
    assert params.top_k == 10
    assert params.max_tokens == 50
    assert params.repetition_penalty == 1.5

    # Test valid params with default values
    params = SamplingParams()
    assert params.temperature == 1.0
    assert params.top_p == 1.0
    assert params.top_k is None
    assert params.max_tokens is None
    assert params.repetition_penalty == 1.0


def test_invalid_temperature():
    """Test invalid temperature values."""
    with pytest.raises(ValueError):
        SamplingParams(temperature=-1.0)


def test_invalid_top_p():
    """Test invalid top_p values."""
    with pytest.raises(ValueError):
        SamplingParams(top_p=0.0)
    with pytest.raises(ValueError):
        SamplingParams(top_p=1.1)


def test_invalid_top_k():
    """Test invalid top_k values."""
    with pytest.raises(ValueError):
        SamplingParams(top_k=0)
    with pytest.raises(ValueError):
        SamplingParams(top_k=-2)


def test_invalid_max_tokens():
    """Test invalid max_tokens values."""
    with pytest.raises(ValueError):
        SamplingParams(max_tokens=0)


def test_kwargs():
    """Test extra keyword arguments."""
    params = SamplingParams(
        temperature=0.5, kwargs={"presence_penalty": 2.0, "frequency_penalty": 1.0}
    )
    assert params.kwargs == {"presence_penalty": 2.0, "frequency_penalty": 1.0}
    assert params.temperature == 0.5
    assert params.top_p == 1.0
    assert params.top_k is None
    assert params.max_tokens is None
    assert params.repetition_penalty == 1.0



# ruff: noqa: S101, NPY002 TODO
import numpy as np
import pytest

from aana.processors.batch import BatchProcessor

NUM_IMAGES = 5
IMAGE_SIZE = 100
FEATURE_SIZE = 10


@pytest.fixture
def images():
    """Gets random-data "images" for use in tests."""
    return np.array([np.random.rand(IMAGE_SIZE, IMAGE_SIZE) for _ in range(NUM_IMAGES)])


@pytest.fixture
def texts():
    """Gets random-data "texts" for use in tests."""
    return [f"text{i}" for i in range(NUM_IMAGES)]


@pytest.fixture
def features():
    """Gets random data "feature vectors" for use in tests."""
    return np.random.rand(NUM_IMAGES, FEATURE_SIZE)


@pytest.fixture
def request_batch(images, texts, features):
    """Sample request batch for use in tests."""
    return {"images": images, "texts": texts, "features": features}


@pytest.fixture
def process_batch():
    """Dummy processing function that just returns the batch."""
    return lambda batch: batch


def test_batch_iterator(request_batch, process_batch):
    """Test batch iterator."""
    batch_size = 2
    processor = BatchProcessor(
        process_batch=process_batch, batch_size=batch_size, num_threads=2
    )

    batches = list(processor.batch_iterator(request_batch))
    # We expect 3 batches with a batch_size of 2 for 5 items
    assert len(batches) == NUM_IMAGES // batch_size + 1
    assert all(
        len(batch["texts"]) == batch_size for batch in batches[:-1]
    )  # All but the last should have batch_size items
    assert len(batches[-1]["texts"]) == 1  # Last batch should have the remaining item


@pytest.mark.asyncio
async def test_process_batches(request_batch, process_batch):
    """Test processing of batches."""
    batch_size = 2
    processor = BatchProcessor(
        process_batch=process_batch, batch_size=batch_size, num_threads=2
    )

    result = await processor.process(request_batch)
    # Ensure all texts are processed
    assert len(result["texts"]) == NUM_IMAGES
    # Check if images are concatenated properly
    assert result["images"].shape == (NUM_IMAGES, IMAGE_SIZE, IMAGE_SIZE)
    # Check if features are concatenated properly
    assert result["features"].shape == (NUM_IMAGES, FEATURE_SIZE)


def test_merge_outputs(request_batch, process_batch):
    """Test merging of outputs from multiple batches.

    images and features should be concatenated into numpy arrays because they are numpy arrays.
    texts should be concatenated because they are lists.
    """
    batch_size = 2
    processor = BatchProcessor(
        process_batch=process_batch, batch_size=batch_size, num_threads=2
    )

    # Assume the processor has already batched and processed the data
    processed_batches = [
        {
            "images": request_batch["images"][:2],
            "texts": request_batch["texts"][:2],
            "features": request_batch["features"][:2],
        },
        {
            "images": request_batch["images"][2:4],
            "texts": request_batch["texts"][2:4],
            "features": request_batch["features"][2:4],
        },
        {
            "images": request_batch["images"][4:],
            "texts": request_batch["texts"][4:],
            "features": request_batch["features"][4:],
        },
    ]

    merged_output = processor.merge_outputs(processed_batches)
    assert merged_output["texts"] == request_batch["texts"]
    assert np.array_equal(merged_output["images"], request_batch["images"])
    assert np.array_equal(merged_output["features"], request_batch["features"])



from abc import ABC, abstractmethod


class EventHandler(ABC):
    """Base class for event handlers. Not guaranteed to be thread safe."""

    @abstractmethod
    def handle(self, event_name: str, *args, **kwargs):
        """Handles an event of the given name.

        Arguments:
            event_name (str): name of the event to handle
            *args (list): specific, context-dependent args
            **kwargs (dict): specific, context-dependent args
        """
        pass



# ruff: noqa: S101
from importlib import resources
from pathlib import Path

import pytest

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_to_dict
from aana.core.models.vad import VadParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.vad_deployment import VadConfig, VadDeployment
from aana.tests.utils import verify_deployment_results

deployments = [
    (
        "vad_deployment",
        VadDeployment.options(
            num_replicas=1,
            max_ongoing_requests=1000,
            ray_actor_options={"num_gpus": 0},
            user_config=VadConfig(
                model=(
                    "https://whisperx.s3.eu-west-2.amazonaws.com/model_weights/segmentation/"
                    "0b5b3216d60a2d32fc086b47ea8c67589aaeb26b7e07fcbe620d6d0b83e209ea/pytorch_model.bin"
                ),
                onset=0.5,
                sample_rate=16000,
            ).model_dump(mode="json"),
        ),
    )
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestVadDeployment:
    """Test VAD deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("audio_file", ["physicsworks.wav", "squirrel.wav"])
    async def test_vad(self, setup_deployment, audio_file):
        """Test VAD."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        audio_file_name = Path(audio_file).stem
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "vad"
            / f"{audio_file_name}.json"
        )

        path = resources.files("aana.tests.files.audios") / audio_file
        assert path.exists(), f"Audio not found: {path}"

        audio = Audio(path=path)

        output = await handle.asr_preprocess_vad(audio=audio, params=VadParams())
        output = pydantic_to_dict(output)
        verify_deployment_results(expected_output_path, output)



# ruff: noqa: F401
# We need to import all db models here and, other than in the class definitions
# themselves, only import them from aana.models.db directly. The reason for
# this is the way SQLAlchemy's declarative base works. You can use forward
# references like `parent = reference("Parent", backreferences="child")`, but the
# forward reference needs to have been resolved before the first constructor
# is called so that SqlAlchemy "knows" about it.
# See:
# https://docs.pylonsproject.org/projects/pyramid_cookbook/en/latest/database/sqlalchemy.html#importing-all-sqlalchemy-models
# (even if not using Pyramid)

from aana.storage.models.base import BaseEntity
from aana.storage.models.caption import CaptionEntity
from aana.storage.models.media import MediaEntity
from aana.storage.models.task import TaskEntity
from aana.storage.models.transcript import TranscriptEntity
from aana.storage.models.video import VideoEntity

__all__ = [
    "BaseEntity",
    "MediaEntity",
    "VideoEntity",
    "CaptionEntity",
    "TranscriptEntity",
]



from typing import Any

import torch
import transformers
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from ray import serve
from transformers import Blip2ForConditionalGeneration, Blip2Processor
from typing_extensions import TypedDict

from aana.core.models.base import pydantic_protected_fields
from aana.core.models.captions import Caption, CaptionsList
from aana.core.models.image import Image
from aana.core.models.types import Dtype
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.processors.batch import BatchProcessor


class HFBlip2Config(BaseModel):
    """The configuration for the BLIP2 deployment with HuggingFace models.

    Attributes:
        model_id (str): The model ID on HuggingFace.
        dtype (Dtype): The data type. Defaults to Dtype.AUTO.
        batch_size (int): The batch size. Defaults to 1.
        num_processing_threads (int): The number of processing threads. Defaults to 1.
        max_new_tokens (int): The maximum numbers of tokens to generate. Defaults to 64.
    """

    model_id: str = Field(validation_alias=AliasChoices("model_id", "model"))
    dtype: Dtype = Field(default=Dtype.AUTO)
    batch_size: int = Field(default=1)
    num_processing_threads: int = Field(default=1)
    max_new_tokens: int = Field(default=64)

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class CaptioningOutput(TypedDict):
    """The output of the captioning model.

    Attributes:
        caption (str): the caption
    """

    caption: Caption


class CaptioningBatchOutput(TypedDict):
    """The output of the captioning model.

    Attributes:
        captions (list[str]): the list of captions
    """

    captions: CaptionsList


@serve.deployment
class HFBlip2Deployment(BaseDeployment):
    """Deployment to serve BLIP2 models using HuggingFace."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and processor from HuggingFace.

        The configuration should conform to the HFBlip2Config schema.
        """
        config_obj = HFBlip2Config(**config)

        # Create the batch processor to split the requests into batches
        # and process them in parallel
        self.batch_size = config_obj.batch_size
        self.num_processing_threads = config_obj.num_processing_threads
        # The actual inference is done in _generate()
        # We use lambda because BatchProcessor expects dict as input
        # and we use **kwargs to unpack the dict into named arguments for _generate()
        self.batch_processor = BatchProcessor(
            process_batch=lambda request: self._generate(**request),
            batch_size=self.batch_size,
            num_threads=self.num_processing_threads,
        )

        # Load the model and processor for BLIP2 from HuggingFace
        self.model_id = config_obj.model_id
        self.dtype = config_obj.dtype
        if self.dtype == Dtype.INT8:
            load_in_8bit = True
            self.torch_dtype = Dtype.FLOAT16.to_torch()
        else:
            load_in_8bit = False
            self.torch_dtype = self.dtype.to_torch()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.max_new_tokens = config_obj.max_new_tokens
        self.model = Blip2ForConditionalGeneration.from_pretrained(
            self.model_id,
            torch_dtype=self.torch_dtype,
            load_in_8bit=load_in_8bit,
            device_map=self.device,
        )
        self.model = torch.compile(self.model)
        self.model.eval()
        self.processor = Blip2Processor.from_pretrained(self.model_id)
        self.model.to(self.device)

    @exception_handler
    async def generate(self, image: Image) -> CaptioningOutput:
        """Generate captions for the given image.

        Args:
            image (Image): the image

        Returns:
            CaptioningOutput: the dictionary with one key "captions"
                            and the list of captions for the image as value

        Raises:
            InferenceException: if the inference fails
        """
        captions: CaptioningBatchOutput = await self.batch_processor.process(
            {"images": [image]}
        )
        return CaptioningOutput(caption=captions["captions"][0])

    @exception_handler
    async def generate_batch(self, **kwargs) -> CaptioningBatchOutput:
        """Generate captions for the given images.

        Args:
            images (List[Image]): the images
            **kwargs (dict[str, Any]): keywordarguments to pass to the
                batch processor.

        Returns:
            CaptioningBatchOutput: the dictionary with one key "captions"
                            and the list of captions for the images as value

        Raises:
            InferenceException: if the inference fails
        """
        # Call the batch processor to process the requests
        # The actual inference is done in _generate()
        return await self.batch_processor.process(kwargs)

    def _generate(self, images: list[Image]) -> CaptioningBatchOutput:
        """Generate captions for the given images.

        This method is called by the batch processor.

        Args:
            images (List[Image]): the images

        Returns:
            CaptioningBatchOutput: the dictionary with one key "captions"
                            and the list of captions for the images as value

        Raises:
            InferenceException: if the inference fails
        """
        # Set the seed to make the results reproducible
        transformers.set_seed(42)
        # Loading images
        numpy_images = [im.get_numpy() for im in images]
        inputs = self.processor(numpy_images, return_tensors="pt").to(
            self.device, self.torch_dtype
        )

        try:
            generated_ids = self.model.generate(
                **inputs, max_new_tokens=self.max_new_tokens
            )
            generated_texts = self.processor.batch_decode(
                generated_ids, skip_special_tokens=True
            )
            generated_texts = [
                generated_text.strip() for generated_text in generated_texts
            ]
            return CaptioningBatchOutput(captions=generated_texts)
        except Exception as e:
            raise InferenceException(self.model_id) from e



# This file is used to define fixtures that are used in the integration tests.
# ruff: noqa: S101
import os
import tempfile
from pathlib import Path
from typing import Any

import portpicker
import pytest
from pytest_postgresql import factories
from sqlalchemy.orm import Session

from aana.configs.db import DbSettings, PostgreSQLConfig, SQLiteConfig
from aana.configs.settings import settings as aana_settings
from aana.sdk import AanaSDK
from aana.storage.op import DbType, run_alembic_migrations
from aana.tests.utils import (
    is_gpu_available,
    send_api_request,
    verify_output,
)
from aana.utils.core import import_from
from aana.utils.json import jsonify

# Change file permission if the user is root
if os.geteuid() == 0:
    postgresql_proc = factories.postgresql_proc(executable="./pg_ctl.sh")
    postgresql = factories.postgresql("postgresql_proc")


@pytest.fixture(scope="module")
def app_factory():
    """Factory fixture to create and configure the app."""

    def create_app(app_module, app_name):
        # Create a temporary database for testing
        tmp_database_path = Path(tempfile.mkstemp(suffix=".db")[1])
        db_config = DbSettings(
            datastore_type=DbType.SQLITE,
            datastore_config=SQLiteConfig(path=tmp_database_path),
        )
        os.environ["DB_CONFIG"] = jsonify(db_config)

        aana_settings.db_config = db_config
        aana_settings.db_config._engine = None

        # Import and start the app
        app = import_from(app_module, app_name)
        app.connect(port=portpicker.pick_unused_port(), show_logs=True, num_cpus=10)
        app.migrate()
        app.deploy()

        return app, tmp_database_path

    return create_app


@pytest.fixture(scope="module")
def create_app():
    """Setup Ray Serve app for given deployments and endpoints."""
    # create temporary database
    tmp_database_path = Path(tempfile.mkstemp(suffix=".db")[1])
    db_config = DbSettings(
        datastore_type=DbType.SQLITE,
        datastore_config=SQLiteConfig(path=tmp_database_path),
    )
    # set environment variable for the database config so Ray can find it
    os.environ["DB_CONFIG"] = jsonify(db_config)

    # set database config in aana settings
    aana_settings.db_config = db_config

    run_alembic_migrations(aana_settings)

    app = AanaSDK()
    app.connect(
        port=portpicker.pick_unused_port(), show_logs=True, num_cpus=10
    )  # pretend we have 10 cpus for testing
    app.migrate()

    def start_app(deployments, endpoints):
        for deployment in deployments:
            deployment_instance = deployment["instance"]

            app.register_deployment(
                name=deployment["name"], instance=deployment_instance
            )

        for endpoint in endpoints:
            app.register_endpoint(
                name=endpoint["name"],
                path=endpoint["path"],
                summary=endpoint["summary"],
                endpoint_cls=endpoint["endpoint_cls"],
                event_handlers=endpoint.get("event_handlers", []),
            )

        app.deploy(blocking=False)

        return app

    yield start_app

    # delete temporary database
    tmp_database_path.unlink()

    app.shutdown()


@pytest.fixture(scope="class")
def setup_deployment(create_app, request):
    """Start the app with provided deployment."""
    deployment_name = request.param[0]
    deployment = request.param[1]

    # skip the test if GPU is not available and the deployment requires GPU
    num_gpus = deployment.ray_actor_options.get("num_gpus", 0)
    if not is_gpu_available() and num_gpus > 0:
        pytest.skip("GPU is not available")

    # keep it the same for all tests so the deployment is replaced
    # when the fixture is called again
    handle_name = "test_deployment"

    deployments = [
        {
            "name": handle_name,
            "instance": deployment,
        }
    ]
    endpoints = []

    return deployment_name, handle_name, create_app(deployments, endpoints)


@pytest.fixture(scope="module")
def call_endpoint(app_setup):
    """Call an endpoint and verify the output."""
    app = app_setup

    def _call_endpoint(
        endpoint_path: str,
        data: dict[str, Any],
        expected_error: str | None = None,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        endpoint = next(
            (e for e in app.endpoints.values() if e.path == endpoint_path), None
        )
        if endpoint is None:
            raise ValueError(f"Endpoint with path {endpoint_path} not found")  # noqa: TRY003
        response = send_api_request(endpoint=endpoint, app=app, data=data)
        verify_output(
            endpoint=endpoint,
            response=response,
            expected_error=expected_error,
        )
        return response

    return _call_endpoint


@pytest.fixture(scope="module")
def one_request_worker():
    """Fixture to update settings to only run one request worker."""
    aana_settings.num_workers = 1
    yield


@pytest.fixture(scope="function")
def sqlite_db_session():
    """Creates a new sql database file and session for each test."""
    with tempfile.NamedTemporaryFile(dir=aana_settings.tmp_data_dir) as tmp:
        # Configure the database to use the temporary file
        aana_settings.db_config.datastore_type = DbType.SQLITE
        aana_settings.db_config.datastore_config = SQLiteConfig(path=tmp.name)
        os.environ["DB_CONFIG"] = jsonify(aana_settings.db_config)

        # Reset the engine
        aana_settings.db_config._engine = None

        run_alembic_migrations(aana_settings)

        # Create a new session
        engine = aana_settings.db_config.get_engine()
        with Session(engine) as session:
            yield session


@pytest.fixture(scope="function")
def postgres_db_session(postgresql):
    """Creates a new postgres database and session for each test."""
    aana_settings.db_config.datastore_type = DbType.POSTGRESQL
    aana_settings.db_config.datastore_config = PostgreSQLConfig(
        host=postgresql.info.host,
        port=postgresql.info.port,
        user=postgresql.info.user,
        password=postgresql.info.password,
        database=postgresql.info.dbname,
    )
    os.environ["DB_CONFIG"] = jsonify(aana_settings.db_config)

    # Reset the engine
    aana_settings.db_config._engine = None

    # Run migrations to set up the schema
    run_alembic_migrations(aana_settings)

    # Create a new session
    engine = aana_settings.db_config.get_engine()
    with Session(engine) as session:
        yield session


@pytest.fixture(params=["sqlite_db_session", "postgres_db_session"])
def db_session(request):
    """Iterate over different database type for db tests."""
    return request.getfixturevalue(request.param)



import inspect
from collections.abc import AsyncGenerator, Callable
from dataclasses import dataclass
from inspect import isasyncgenfunction
from typing import Annotated, Any, get_origin

from fastapi import FastAPI, File, Form, Query, UploadFile
from fastapi.responses import StreamingResponse
from pydantic import ConfigDict, Field, ValidationError, create_model
from pydantic.main import BaseModel

from aana.api.event_handlers.event_handler import EventHandler
from aana.api.event_handlers.event_manager import EventManager
from aana.api.exception_handler import custom_exception_handler
from aana.api.responses import AanaJSONResponse
from aana.configs.settings import settings as aana_settings
from aana.core.models.exception import ExceptionResponseModel
from aana.exceptions.runtime import (
    MultipleFileUploadNotAllowed,
)
from aana.storage.repository.task import TaskRepository
from aana.storage.session import get_session


def get_default_values(func):
    """Get the default values for the function arguments."""
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


@dataclass
class FileUploadField:
    """Class used to represent a file upload field.

    Attributes:
        name (str): Name of the field.
        description (str): Description of the field.
    """

    name: str
    description: str


@dataclass
class Endpoint:
    """Class used to represent an endpoint.

    Attributes:
        name (str): Name of the endpoint.
        path (str): Path of the endpoint (e.g. "/video/transcribe").
        summary (str): Description of the endpoint that will be shown in the API documentation.
        event_handlers (list[EventHandler] | None): The list of event handlers to register for the endpoint.
    """

    name: str
    path: str
    summary: str
    initialized: bool = False
    event_handlers: list[EventHandler] | None = None

    async def initialize(self):
        """Initialize the endpoint.

        Redefine this method to add initialization logic for the endpoint (e.g. create a handle to the deployment).
        Call super().initialize() to ensure the endpoint is initialized.

        Example:
            ```python
            async def initialize(self):
                await super().initialize()
                self.asr_handle = await AanaDeploymentHandle.create("whisper_deployment")
            ```
        """
        self.initialized = True

    async def run(self, *args, **kwargs):
        """The main method of the endpoint that is called when the endpoint receives a request.

        Redefine this method to implement the logic of the endpoint.

        Example:
            ```python
            async def run(self, video: VideoInput) -> WhisperOutput:
                video_obj: Video = await run_remote(download_video)(video_input=video)
                transcription = await self.asr_handle.transcribe(audio=audio)
                return transcription
            ```
        """
        raise NotImplementedError

    def register(
        self, app: FastAPI, custom_schemas: dict[str, dict], event_manager: EventManager
    ):
        """Register the endpoint in the FastAPI application.

        Args:
            app (FastAPI): FastAPI application.
            custom_schemas (dict[str, dict]): Dictionary containing custom schemas.
            event_manager (EventManager): Event manager for the application.
        """
        RequestModel = self.get_request_model()
        ResponseModel = self.get_response_model()

        file_upload_field = self.__get_file_upload_field()
        if self.event_handlers:
            for handler in self.event_handlers:
                event_manager.register_handler_for_events(handler, [self.path])

        route_func = self.__create_endpoint_func(
            RequestModel=RequestModel,
            file_upload_field=file_upload_field,
            event_manager=event_manager,
        )

        app.post(
            self.path,
            name=self.name,
            summary=self.summary,
            operation_id=self.name,
            response_model=ResponseModel,
            responses={
                400: {"model": ExceptionResponseModel},
            },
        )(route_func)
        custom_schemas[self.name] = RequestModel.model_json_schema()

    def __generate_model_name(self, suffix: str) -> str:
        """Generate a Pydantic model name based on a given suffix.

        Args:
            suffix (str): Suffix for the model name (e.g. "Request", "Response").

        Returns:
            str: Generated model name.
        """
        return self.__class__.__name__ + suffix

    def __arg_to_field(self, arg_name: str, arg_type: Any) -> tuple[Any, Any]:
        """Convert an argument to a Pydantic field.

        Args:
            arg_name (str): Name of the argument.
            arg_type (Any): Type of the argument.

        Returns:
            Tuple[Any, Any]: Tuple containing the Pydantic model and the Pydantic field.
        """
        # if data model is None or Any, set it to Any
        if arg_type is None or arg_type == Any:
            return (Any, Field(None))

        if get_origin(arg_type) is Annotated:
            default = get_default_values(self.run).get(arg_name)
            if default is not None:
                return (arg_type, default)
            else:
                return (arg_type, ...)

        # try to instantiate the data model
        # to see if any of the fields are required
        try:
            data_model_instance = arg_type()
        except TypeError as e:
            raise ValueError(  # noqa: TRY003
                f"Invalid type for argument {arg_name}: {arg_type}. "
                "Consider using Annotated[..., Field(...)] or Pydantic models."
            ) from e
        except ValidationError:
            # if we can't instantiate the data model
            # it means that it has required fields
            return (arg_type, ...)
        else:
            return (arg_type, data_model_instance)

    def __get_input_fields(self) -> dict[str, tuple[Any, Any]]:
        """Generate fields for the request Pydantic model based on the function annotations.

        Returns:
            dict[str, tuple[Any, Field]]: Dictionary of fields for the request Pydantic model.
        """
        fields = {}
        for arg_name, arg_type in self.run.__annotations__.items():
            if arg_name == "return":
                continue
            fields[arg_name] = self.__arg_to_field(arg_name, arg_type)
        return fields

    def get_request_model(self) -> type[BaseModel]:
        """Generate the request Pydantic model for the endpoint.

        Returns:
            type[BaseModel]: Request Pydantic model.
        """
        model_name = self.__generate_model_name("Request")
        input_fields = self.__get_input_fields()
        return create_model(model_name, **input_fields)

    def __get_output_fields(self) -> dict[str, tuple[Any, Any]]:
        """Generate fields for the response Pydantic model based on the function annotations.

        Returns:
            dict[str, tuple[Any, Any]]: Dictionary of fields for the response Pydantic model.
        """
        try:
            if self.is_streaming_response():
                return_type = self.run.__annotations__["return"].__args__[0]
            else:
                return_type = self.run.__annotations__["return"]
        except (AttributeError, KeyError) as e:
            raise ValueError("Endpoint function must have a return annotation.") from e  # noqa: TRY003
        fields = {}
        for arg_name, arg_type in return_type.__annotations__.items():
            fields[arg_name] = (arg_type, Field(None))
        return fields

    def get_response_model(self) -> type[BaseModel]:
        """Generate the response Pydantic model for the endpoint.

        Returns:
            type[BaseModel]: Response Pydantic model.
        """
        model_name = self.__generate_model_name("Response")
        output_fields = self.__get_output_fields()
        return create_model(
            model_name, **output_fields, __config__=ConfigDict(extra="forbid")
        )

    def __get_file_upload_field(self) -> FileUploadField | None:
        """Get the file upload field for the endpoint.

        Returns:
            Optional[FileUploadField]: File upload field or None if not found.

        Raises:
            MultipleFileUploadNotAllowed: If multiple inputs require file upload.
        """
        file_upload_field = None
        for arg_name, arg_type in self.run.__annotations__.items():
            if arg_name == "return":
                continue

            # check if pydantic model has file_upload field and it's set to True
            if isinstance(arg_type, type) and issubclass(arg_type, BaseModel):
                file_upload_enabled = arg_type.model_config.get("file_upload", False)
                file_upload_description = arg_type.model_config.get(
                    "file_upload_description", ""
                )
            else:
                file_upload_enabled = False
                file_upload_description = ""

            if file_upload_enabled and file_upload_field is None:
                file_upload_field = FileUploadField(
                    name=arg_name, description=file_upload_description
                )
            elif file_upload_enabled and file_upload_field is not None:
                # raise an exception if multiple inputs require file upload
                raise MultipleFileUploadNotAllowed(arg_name)
        return file_upload_field

    @classmethod
    def is_streaming_response(cls) -> bool:
        """Check if the endpoint returns a streaming response.

        Returns:
            bool: True if the endpoint returns a streaming response, False otherwise.
        """
        return isasyncgenfunction(cls.run)

    def __create_endpoint_func(  # noqa: C901
        self,
        RequestModel: type[BaseModel],
        file_upload_field: FileUploadField | None = None,
        event_manager: EventManager | None = None,
    ) -> Callable:
        """Create a function for routing an endpoint."""
        # Copy path to a bound variable so we don't retain an external reference
        bound_path = self.path

        async def route_func_body(  # noqa: C901
            body: str, files: list[UploadFile] | None = None, defer=False
        ):
            if not self.initialized:
                await self.initialize()

            if event_manager:
                event_manager.handle(bound_path, defer=defer)

            # parse form data as a pydantic model and validate it
            data = RequestModel.model_validate_json(body)

            # if the input requires file upload, add the files to the data
            if file_upload_field and files:
                files_as_bytes = [await file.read() for file in files]
                getattr(data, file_upload_field.name).set_files(files_as_bytes)

            # We have to do this instead of data.dict() because
            # data.dict() will convert all nested models to dicts
            # and we want to keep them as pydantic models
            data_dict = {}
            for field_name in data.model_fields:
                field_value = getattr(data, field_name)
                data_dict[field_name] = field_value

            if defer:
                if not aana_settings.task_queue.enabled:
                    raise RuntimeError("Task queue is not enabled.")  # noqa: TRY003

                with get_session() as session:
                    task = TaskRepository(session).save(
                        endpoint=bound_path, data=data_dict
                    )
                    return AanaJSONResponse(content={"task_id": str(task.id)})

            if isasyncgenfunction(self.run):

                async def generator_wrapper() -> AsyncGenerator[bytes, None]:
                    """Serializes the output of the generator using ORJSONResponseCustom."""
                    try:
                        async for output in self.run(**data_dict):
                            yield AanaJSONResponse(content=output).body
                    except Exception as e:
                        yield custom_exception_handler(None, e).body

                return StreamingResponse(
                    generator_wrapper(), media_type="application/json"
                )
            else:
                try:
                    output = await self.run(**data_dict)
                except Exception as e:
                    return custom_exception_handler(None, e)
                return AanaJSONResponse(content=output)

        if file_upload_field:
            files = File(None, description=file_upload_field.description)
        else:
            files = None

        async def route_func(
            body: str = Form(...),
            files=files,
            defer: bool = Query(
                description="Defer execution of the endpoint to the task queue.",
                default=False,
                include_in_schema=aana_settings.task_queue.enabled,
            ),
        ):
            return await route_func_body(body=body, files=files, defer=defer)

        return route_func


def add_custom_schemas_to_openapi_schema(
    openapi_schema: dict[str, Any], custom_schemas: dict[str, Any]
) -> dict[str, Any]:
    """Add custom schemas to the openapi schema.

    File upload is that FastAPI doesn't support Pydantic models in multipart requests.
    There is a discussion about it on FastAPI discussion forum.
    See https://github.com/tiangolo/fastapi/discussions/8406
    The topic starter suggests a workaround.
    The workaround is to use Forms instead of Pydantic models in the endpoint definition and
    then convert the Forms to Pydantic models in the endpoint itself
    using parse_raw_as function from Pydantic.
    Since Pydantic model isn't used in the endpoint definition,
    the API documentation will not be generated automatically.
    So the workaround also suggests updating the API documentation manually
    by overriding the openapi method of a FastAPI application.

    Args:
        openapi_schema (dict): The openapi schema.
        custom_schemas (dict): The custom schemas.

    Returns:
        dict: The openapi schema with the custom schemas added.
    """
    if "$defs" not in openapi_schema:
        openapi_schema["$defs"] = {}
    for schema_name, schema in custom_schemas.items():
        # if we have a definitions then we need to move them out to the top level of the schema
        if "$defs" in schema:
            openapi_schema["$defs"].update(schema["$defs"])
            del schema["$defs"]
        openapi_schema["components"]["schemas"][f"Body_{schema_name}"]["properties"][
            "body"
        ] = schema
    return openapi_schema



from pathlib import Path
from typing import Any

import orjson
from pydantic import BaseModel
from sqlalchemy import Engine

__all__ = ["jsonify", "json_serializer_default"]


def json_serializer_default(obj: object) -> object:
    """Default function for json serializer to handle custom objects.

    If json serializer does not know how to serialize an object, it calls the default function.

    For example, if we see that the object is a pydantic model,
    we call the dict method to get the dictionary representation of the model
    that json serializer can deal with.

    If the object is not supported, we raise a TypeError.

    Args:
        obj (object): The object to serialize.

    Returns:
        object: The serializable object.

    Raises:
        TypeError: If the object is not a pydantic model, Path, or Media object.
    """
    if isinstance(obj, Engine):
        return None
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, type):
        return str(type)

    from aana.core.models.media import Media
    if isinstance(obj, Media):
        return str(obj)

    raise TypeError(type(obj))


def jsonify(data: Any, option: int | None = orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS, as_bytes: bool = False) -> str | bytes:
    """Serialize content using orjson.

    Args:
        data (Any): The content to serialize.
        option (int | None): The option for orjson.dumps.
        as_bytes (bool): Return output as bytes instead of string

    Returns:
        bytes | str: The serialized data as desired format.
    """
    output = orjson.dumps(data, option=option, default=json_serializer_default)
    return output if as_bytes else output.decode()



# ruff: noqa: S101

from aana.core.models.chat import Question


def test_question_creation():
    """Test that a question can be created."""
    question = Question("What is the capital of France?")
    assert question == "What is the capital of France?"



import hashlib
from pathlib import Path

import yt_dlp
from yt_dlp.utils import DownloadError

from aana.configs.settings import settings
from aana.core.models.video import Video, VideoInput, VideoMetadata
from aana.exceptions.io import (
    DownloadException,
)

__all__ = ["download_video", "get_video_metadata"]


def get_video_metadata(video_url: str) -> VideoMetadata:
    """Fetch video's metadata for a url.

    Args:
        video_url (str): the video input url

    Returns:
        metadata (VideoMetadata): the metadata of the video

    Raises:
        DownloadException: Request does not succeed.
    """
    ydl_options = {
        "extract_flat": True,
        "hls_prefer_native": True,
        "extractor_args": {"youtube": {"skip": ["hls", "dash"]}},
    }
    try:
        with yt_dlp.YoutubeDL(ydl_options) as ydl:
            info = ydl.extract_info(video_url, download=False)
            title = info.get("title", "")
            description = info.get("description", "")
            duration = info.get("duration")
            return VideoMetadata(
                title=title,
                description=description,
                duration=duration,
            )
    except DownloadError as e:
        error_message = e.msg.split(";")[0]
        raise DownloadException(url=video_url, msg=error_message) from e


def download_video(video_input: VideoInput | Video) -> Video:
    """Downloads videos for a VideoInput object.

    Args:
        video_input (VideoInput): the video input to download

    Returns:
        Video: the video object

    Raises:
        DownloadException: Request does not succeed.
    """
    if isinstance(video_input, Video):
        return video_input
    if video_input.url is not None:
        video_dir = settings.video_dir
        url_hash = hashlib.md5(
            video_input.url.encode(), usedforsecurity=False
        ).hexdigest()

        # we use yt_dlp to download the video
        # it works not only for youtube videos, but also for other websites and direct links
        ydl_options = {
            "outtmpl": f"{video_dir}/{url_hash}.%(ext)s",
            "extract_flat": True,
            "hls_prefer_native": True,
            "extractor_args": {"youtube": {"skip": ["hls", "dash"]}},
        }
        try:
            with yt_dlp.YoutubeDL(ydl_options) as ydl:
                info = ydl.extract_info(video_input.url, download=False)
                title = info.get("title", "")
                description = info.get("description", "")
                path = Path(ydl.prepare_filename(info))
                if not path.exists():
                    ydl.download([video_input.url])
                if not path.exists():
                    raise DownloadException(video_input.url)
                return Video(
                    path=path,
                    url=video_input.url,
                    media_id=video_input.media_id,
                    title=title,
                    description=description,
                )
        except DownloadError as e:
            # removes the yt-dlp request to file an issue
            error_message = e.msg.split(";")[0]
            raise DownloadException(url=video_input.url, msg=error_message) from e
    else:
        return video_input.convert_input_to_object()



from collections.abc import AsyncGenerator
from threading import Thread
from typing import Any

import torch
import transformers
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from ray import serve
from transformers import (
    AutoModelForVision2Seq,
    AutoProcessor,
    TextIteratorStreamer,
)
from transformers.utils.import_utils import is_flash_attn_2_available

from aana.core.models.base import merged_options, pydantic_protected_fields
from aana.core.models.chat import ChatMessage
from aana.core.models.custom_config import CustomConfig
from aana.core.models.image_chat import ImageChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.deployments.base_text_generation_deployment import ChatOutput, LLMOutput
from aana.exceptions.runtime import InferenceException
from aana.utils.streamer import async_streamer_adapter


class Idefics2Config(BaseModel):
    """The configuration for the Idefics 2 model.

    Attributes:
        model_id (str): The model ID on HuggingFace.
        dtype (Dtype): The data type. Defaults to Dtype.AUTO.
        enable_flash_attention_2 (bool | None): Use Flash Attention 2. If None, Flash Attention 2 wii be enabled if available. Defaults to None.
        model_kwargs (CustomConfig): The extra model keyword arguments. Defaults to {}.
        processor_kwargs (CustomConfig): The extra processor keyword arguments. Defaults to {}.
        default_sampling_params (SamplingParams): The default sampling parameters. Defaults to SamplingParams(temperature=1.0, max_tokens=256).
    """

    model_id: str = Field(validation_alias=AliasChoices("model_id", "model"))
    model_kwargs: CustomConfig = {}
    processor_kwargs: CustomConfig = {}
    enable_flash_attention_2: bool | None = Field(default=None)
    dtype: Dtype = Field(default=Dtype.AUTO)
    default_sampling_params: SamplingParams = SamplingParams(
        temperature=1.0, max_tokens=256
    )
    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment
class Idefics2Deployment(BaseDeployment):
    """Deployment to serve the Idefics 2 model."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and processor from HuggingFace.

        The configuration should conform to the Idefics2Config schema.
        """
        config_obj = Idefics2Config(**config)

        self.model_id = config_obj.model_id
        self.model_kwargs = config_obj.model_kwargs
        self.dtype = config_obj.dtype
        self.default_sampling_params = config_obj.default_sampling_params
        self.torch_dtype = self.dtype.to_torch()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        self.processor = AutoProcessor.from_pretrained(
            self.model_id, **config_obj.processor_kwargs
        )
        self.model_kwargs.update(
            dict(
                torch_dtype=self.torch_dtype,
                device_map=self.device,
            )
        )

        if config_obj.enable_flash_attention_2 is None:
            config_obj.enable_flash_attention_2 = is_flash_attn_2_available()
        if config_obj.enable_flash_attention_2:
            self.model_kwargs["_attn_implementation"] = "flash_attention_2"
        self.model = AutoModelForVision2Seq.from_pretrained(
            self.model_id, **self.model_kwargs
        )

    @exception_handler
    async def chat_stream(
        self, dialog: ImageChatDialog, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Chat with the vision model and stream the results.

        Args:
            dialog (ImageChatDialog): the dialog with images
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the chunk of generated text as the value
        """
        # Set the seed to make the results reproducible
        transformers.set_seed(42)

        if sampling_params is None:
            sampling_params = self.default_sampling_params
        else:
            sampling_params = merged_options(
                self.default_sampling_params, sampling_params
            )

        messages, images = dialog.to_objects()
        text = self.processor.apply_chat_template(messages, add_generation_prompt=True)
        inputs = self.processor(
            images=[img.get_pil_image() for img in images],
            text=text,
            return_tensors="pt",
        ).to(self.device)

        try:
            with torch.no_grad():
                streamer = TextIteratorStreamer(
                    self.processor.tokenizer,
                    timeout=0,
                    skip_prompt=True,
                    skip_special_tokens=True,
                )
                generation_kwargs = dict(
                    **inputs,
                    streamer=streamer,
                    max_new_tokens=sampling_params.max_tokens,
                    top_k=sampling_params.top_k,
                    top_p=sampling_params.top_p,
                    temperature=sampling_params.temperature,
                    num_return_sequences=1,
                    eos_token_id=self.processor.tokenizer.eos_token_id,
                    **sampling_params.kwargs,
                )
                if sampling_params.temperature == 0:
                    generation_kwargs["do_sample"] = False
                    generation_kwargs["temperature"] = None
                generation_thread = Thread(
                    target=self.model.generate, kwargs=generation_kwargs
                )
                generation_thread.start()

                async_streamer = async_streamer_adapter(streamer)
                async for new_text in async_streamer:
                    yield LLMOutput(text=new_text)

                # clean up
                inputs.to("cpu")
                del streamer
                del inputs
        except Exception as e:
            raise InferenceException(model_name=self.model_id) from e

    async def chat(
        self, dialog: ImageChatDialog, sampling_params: SamplingParams | None = None
    ) -> ChatOutput:
        """Chat with the vision model.

        Args:
            dialog (ImageChatDialog): the dialog with images
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            ChatOutput: the chat output with the message
        """
        text = ""
        async for chunk in self.chat_stream(dialog, sampling_params):
            text += chunk["text"]

        return ChatOutput(message=ChatMessage(content=text, role="assistant"))

    @exception_handler
    async def chat_batch(
        self,
        dialogs: list[ImageChatDialog],
        sampling_params: SamplingParams | None = None,
    ) -> list[ChatOutput]:
        """Chat with the vision model in batch mode.

        Args:
            dialogs (list[ImageChatDialog]): the list of dialogs with images
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            list[ChatOutput]: the list of chat outputs with the messages
        """
        # Set the seed to make the results reproducible
        transformers.set_seed(42)

        if sampling_params is None:
            sampling_params = self.default_sampling_params
        else:
            sampling_params = merged_options(
                self.default_sampling_params, sampling_params
            )

        text_batch = []
        image_batch = []
        for dialog in dialogs:
            messages, images = dialog.to_objects()
            text = self.processor.apply_chat_template(
                messages, add_generation_prompt=True
            )
            text_batch.append(text)
            image_batch.append([img.get_pil_image() for img in images])

        inputs = self.processor(
            images=image_batch,
            text=text_batch,
            return_tensors="pt",
            padding=True,
        ).to(self.device)

        try:
            generation_kwargs = dict(
                **inputs,
                max_new_tokens=sampling_params.max_tokens,
                top_k=sampling_params.top_k,
                top_p=sampling_params.top_p,
                temperature=sampling_params.temperature,
                num_return_sequences=1,
                eos_token_id=self.processor.tokenizer.eos_token_id,
            )
            if sampling_params.temperature == 0:
                generation_kwargs["do_sample"] = False
                generation_kwargs["temperature"] = None

            generated_texts = self.model.generate(**generation_kwargs)
            generated_texts = self.processor.batch_decode(
                generated_texts, skip_special_tokens=True
            )

            chat_outputs = []
            for text in generated_texts:
                # Remove the prompt from the generated text
                # TODO: find a better way to remove the prompt, it will not work for other prompt formats
                text = text.split("Assistant:")[1].strip()
                chat_outputs.append(
                    ChatOutput(message=ChatMessage(content=text, role="assistant"))
                )
        except Exception as e:
            raise InferenceException(model_name=self.model_id) from e
        else:
            return chat_outputs



import gc
import io
import itertools
import wave
from collections.abc import Generator
from pathlib import Path

import av
import numpy as np

from aana.core.libraries.audio import AbstractAudioLibrary

__all__ = ["load_audio", "pyAVWrapper"]


def load_audio(file: Path, sample_rate: int = 16000) -> bytes:
    """Open an audio file and read as mono waveform, resampling as necessary.

    Args:
        file (Path): The audio/video file to open.
        sample_rate (int): The sample rate to resample the audio if necessary.

    Returns:
        bytes: The content of the audio as bytes.

    Raises:
        RuntimeError: if ffmpeg fails to convert and load the audio.
    """
    resampler = av.audio.resampler.AudioResampler(
        format="s16",
        layout="mono",
        rate=sample_rate,
    )

    raw_buffer = io.BytesIO()

    # Try loading audio and check for empty audio in one shot.
    try:
        with av.open(str(file), mode="r", metadata_errors="ignore") as container:
            # check for empty audio
            if container.streams.audio == tuple():
                return b""

            frames = container.decode(audio=0)
            frames = ignore_invalid_frames(frames)
            frames = group_frames(frames, 500000)
            frames = resample_frames(frames, resampler)

            for frame in frames:
                array = frame.to_ndarray()
                raw_buffer.write(array)

        # It appears that some objects related to the resampler are not freed
        # unless the garbage collector is manually run.
        del resampler
        gc.collect()

        return raw_buffer.getvalue()

    except Exception as e:
        raise RuntimeError(f"{e!s}") from e


def ignore_invalid_frames(frames: Generator) -> Generator:
    """Filter out invalid frames from the input generator.

    Args:
        frames (Generator): The input generator of frames.

    Yields:
        av.audio.frame.AudioFrame: Valid audio frames.

    Raises:
        StopIteration: When the input generator is exhausted.
    """
    iterator = iter(frames)

    while True:
        try:
            yield next(iterator)
        except StopIteration:  # noqa: PERF203
            break
        except av.error.InvalidDataError:
            continue


def group_frames(frames: Generator, num_samples: int | None = None) -> Generator:
    """Group audio frames and yield groups of frames based on the specified number of samples.

    Args:
        frames (Generator): The input generator of audio frames.
        num_samples (int | None): The target number of samples for each group.

    Yields:
        av.audio.frame.AudioFrame: Grouped audio frames.
    """
    fifo = av.audio.fifo.AudioFifo()

    for frame in frames:
        frame.pts = None  # Ignore timestamp check.
        fifo.write(frame)

        if num_samples is not None and fifo.samples >= num_samples:
            yield fifo.read()

    if fifo.samples > 0:
        yield fifo.read()


def resample_frames(frames: Generator, resampler) -> Generator:
    """Resample audio frames using the provided resampler.

    Args:
        frames (Generator): The input generator of audio frames.
        resampler: The audio resampler.

    Yields:
        av.audio.frame.AudioFrame: Resampled audio frames.
    """
    # Add None to flush the resampler.
    for frame in itertools.chain(frames, [None]):
        yield from resampler.resample(frame)


class pyAVWrapper(AbstractAudioLibrary):
    """Class for audio handling using PyAV library."""

    @classmethod
    def read_file(cls, path: Path, sample_rate: int = 16000) -> np.ndarray:
        """Read an audio file from path and return it as a numpy array.

        Args:
            path (Path): The path of the file to read.
            sample_rate (int): sample rate of the audio, default is 16000.

        Returns:
            np.ndarray: The audio file as a numpy array.
        """
        resampler = av.audio.resampler.AudioResampler(
            format="s16",
            layout="mono",
            rate=sample_rate,
        )

        raw_buffer = io.BytesIO()
        dtype = None

        with av.open(str(path), mode="r", metadata_errors="ignore") as container:
            frames = container.decode(audio=0)
            frames = ignore_invalid_frames(frames)
            frames = group_frames(frames, 500000)
            frames = resample_frames(frames, resampler)

            for frame in frames:
                array = frame.to_ndarray()
                dtype = array.dtype
                raw_buffer.write(array)

        # It appears that some objects related to the resampler are not freed
        # unless the garbage collector is manually run.
        del resampler
        gc.collect()

        audio = np.frombuffer(raw_buffer.getbuffer(), dtype=dtype)
        # Convert s16 back to f32.
        audio = audio.astype(np.float32) / 32768.0
        return audio

    @classmethod
    def read_from_bytes(cls, content: bytes, sample_rate: int = 16000) -> np.ndarray:
        """Read audio bytes and return as a numpy array.

        Args:
            content (bytes): The content of the file to read.
            sample_rate (int): sample rate of the audio, default is 16000.

        Returns:
            np.ndarray: The file as a numpy array.
        """
        # Create an in-memory file-like object
        content_io = io.BytesIO(content)

        resampler = av.audio.resampler.AudioResampler(
            format="s16",
            layout="mono",
            rate=sample_rate,
        )

        raw_buffer = io.BytesIO()
        dtype = None

        with av.open(content_io, mode="r", metadata_errors="ignore") as container:
            frames = container.decode(audio=0)
            frames = ignore_invalid_frames(frames)
            frames = group_frames(frames, 500000)
            frames = resample_frames(frames, resampler)

            for frame in frames:
                array = frame.to_ndarray()
                dtype = array.dtype
                raw_buffer.write(array)

        # It appears that some objects related to the resampler are not freed
        # unless the garbage collector is manually run.
        del resampler
        gc.collect()

        audio = np.frombuffer(raw_buffer.getbuffer(), dtype=dtype)
        # Convert s16 back to f32.
        audio = audio.astype(np.float32) / 32768.0
        return audio

    @classmethod
    def write_file(cls, path: Path, audio: np.ndarray, sample_rate: int = 16000):
        """Write an audio file in wav format to the path from numpy array.

        Args:
            path (Path): The path of the file to write.
            audio (np.ndarray): The audio to write.
            sample_rate (int): The sample rate of the audio to save, default is 16000.
        """
        audio = (audio * 32768.0).astype(np.int16)
        # Create an AV container
        container = av.open(str(path), "w", format="wav")
        # Add an audio stream
        stream = container.add_stream("pcm_s16le", rate=sample_rate)
        stream.channels = 1
        # Write audio frames to the stream
        for frame in av.AudioFrame.from_ndarray(
            audio, format="s16", layout="mono", rate=sample_rate
        ):
            for packet in stream.encode(frame):
                container.mux(packet)
        for packet in stream.encode(None):
            container.mux(packet)
        container.close()

    @classmethod
    def write_to_bytes(cls, audio: np.ndarray) -> bytes:
        """Write bytes using the audio library from numpy array.

        Args:
            audio (np.ndarray): The audio to write.

        Returns:
            bytes: The audio as bytes.
        """
        frame = av.AudioFrame(format="s16", layout="mono", samples=len(audio))
        frame.planes[0].update(audio.astype(np.int16).tobytes())
        return frame.planes[0].to_bytes()

    @classmethod
    def write_audio_bytes(cls, path: Path, audio: bytes, sample_rate: int = 16000):
        """Write an audio file in wav format to path from the normalized audio bytes.

        Args:
            path (Path): The path of the file to write.
            audio (bytes): The audio to in 16-bit PCM byte write.
            sample_rate (int): The sample rate of the audio, default is 16000.
        """
        with wave.open(str(path), "wb") as wav_file:
            wav_file.setnchannels(1)  # Mono audio
            wav_file.setsampwidth(2)  # 16-bit audio
            wav_file.setframerate(sample_rate)  # Sample rate
            wav_file.writeframes(audio)



from typing import Any, TypedDict

import torch
from huggingface_hub.utils import GatedRepoError
from pyannote.audio import Pipeline
from pyannote.core import Annotation
from pydantic import BaseModel, ConfigDict, Field
from ray import serve

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_protected_fields
from aana.core.models.speaker import (
    PyannoteSpeakerDiarizationParams,
    SpeakerDiarizationSegment,
)
from aana.core.models.time import TimeInterval
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.processors.speaker import combine_homogeneous_speaker_diarization_segments


class SpeakerDiarizationOutput(TypedDict):
    """The output of the Speaker Diarization model.

    Attributes:
        segments (list[SpeakerDiarizationSegment]): The Speaker Diarization segments.
    """

    segments: list[SpeakerDiarizationSegment]


class PyannoteSpeakerDiarizationConfig(BaseModel):
    """The configuration for the Pyannote Speaker Diarization deployment.

    Attributes:
        model_id (str): name of the speaker diarization pipeline.
        sample_rate (int): The sample rate of the audio. Defaults to 16000.
    """

    model_id: str = Field(
        default="pyannote/speaker-diarization-3.1",
        description="The Speaker Diarization model ID.",
    )

    sample_rate: int = Field(default=16000, description="Sample rate of the audio.")

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment(health_check_timeout_s=180)
class PyannoteSpeakerDiarizationDeployment(BaseDeployment):
    """Deployment to serve Pyannote Speaker Diarization (SD) models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and instantiate speaker diarization pipeline.

        The configuration should conform to the PyannoteSpeakerDiarizationConfig schema.

        """
        config_obj = PyannoteSpeakerDiarizationConfig(**config)

        self.sample_rate = config_obj.sample_rate
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model_id = config_obj.model_id

        # for consistency across multiple runs
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(42)

        try:
            # load model using pyannote Pipeline
            self.diarize_model = Pipeline.from_pretrained(self.model_id)

            if self.diarize_model:
                self.diarize_model.to(torch.device(self.device))

        except Exception as e:
            raise GatedRepoError(
                message=f"This repository is private and requires a token to accept user conditions and access models in {self.model_id} pipeline."
            ) from e

    async def __inference(
        self, audio: Audio, params: PyannoteSpeakerDiarizationParams
    ) -> Annotation:
        """Perform inference on the Audio with the Pyannote Speaker Diarization model.

        Args:
            audio (Audio): The audio to perform Speaker Diarization.
            params (PyannoteSpeakerDiarizationParams): Parameters for the pyannote speaker diarization model.

        Returns:
            speaker_segments (Annotation): The list of speaker diarized segments.

        Raises:
            InferenceException: If the Speaker Diarization inference fails.
        """
        audio_array = audio.get_numpy()
        speaker_diarization_input = {
            "waveform": torch.from_numpy(audio_array).unsqueeze(0),
            "sample_rate": self.sample_rate,
        }

        try:
            speaker_segments = self.diarize_model(
                speaker_diarization_input,
                min_speakers=params.min_speakers,
                max_speakers=params.max_speakers,
            )

        except Exception as e:
            raise InferenceException(self.model_id) from e

        return speaker_segments

    @exception_handler
    async def diarize(
        self, audio: Audio, params: PyannoteSpeakerDiarizationParams | None = None
    ) -> SpeakerDiarizationOutput:
        """Perform Speaker Diarization inference to get speaker segments.

        Args:
            audio (Audio): The audio to perform Speaker Diarization.
            params (PyannoteSpeakerDiarizationParams): Parameters for the speaker diarization model.

        Returns:
            SpeakerDiarizationOutput: Output speaker segments from the Speaker Diarization pipeline.

        """
        if not params:
            params = PyannoteSpeakerDiarizationParams()

        speaker_segments = await self.__inference(audio, params)
        speaker_diarization_segments = []
        for speech_turn, _, speaker in speaker_segments.itertracks(yield_label=True):
            speaker_diarization_segments.append(
                SpeakerDiarizationSegment(
                    time_interval=TimeInterval(
                        start=speech_turn.start, end=speech_turn.end
                    ),
                    speaker=speaker,
                )
            )

        # Combine homogeneous speaker segments.
        processed_speaker_diarization_segments = (
            combine_homogeneous_speaker_diarization_segments(
                speaker_diarization_segments
            )
        )
        return SpeakerDiarizationOutput(segments=processed_speaker_diarization_segments)



import hashlib
import io
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import numpy as np
import PIL.Image
from pydantic import (
    AfterValidator,
    AnyUrl,
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)
from pydantic_core import InitErrorDetails
from typing_extensions import Self

from aana.configs.settings import settings
from aana.core.libraries.image import AbstractImageLibrary
from aana.core.models.base import BaseListModel
from aana.core.models.media import Media, MediaId
from aana.exceptions.io import ImageReadingException
from aana.integrations.external.opencv import OpenCVWrapper
from aana.utils.download import download_file


@dataclass
class Image(Media):
    """A class representing an image.

    At least one of `path`, `url`, `content` or `numpy` must be provided.
    If `save_on_disk` is True, the image will be saved on disk automatically.

    Attributes:
        path (Path): The file path of the image.
        url (str): The URL of the image.
        content (bytes): The content of the image in bytes (image file as bytes).
        numpy (np.ndarray): The image as a numpy array.
        media_id (MediaId): The ID of the image, generated automatically if not provided.
    """

    media_dir: Path | None = settings.image_dir
    numpy: np.ndarray | None = None  # The image as a numpy array.
    image_lib: type[
        AbstractImageLibrary
    ] = OpenCVWrapper  # The image library to use, TODO: add support for PIL and allow to choose the library

    def _validate(self):
        """Validate the image."""
        # validate the parent class
        super()._validate()

        # check that at least one of 'path', 'url', 'content' or 'numpy' is provided
        if not any(
            [
                self.path is not None,
                self.url is not None,
                self.content is not None,
                self.numpy is not None,
            ]
        ):
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )

    def save(self):
        """Save the image on disk.

        If the image is already available on disk, do nothing.
        If the image represented as a byte string, save it on disk.
        If the image is represented as a URL, download it and save it on disk.
        If the image is represented as a numpy array, convert it to BMP and save it on disk.

        First check if the image is already available on disk, then content, then url, then numpy
        to avoid unnecessary operations (e.g. downloading the image or converting it to BMP).

        Raises:
            ValueError: If none of 'path', 'url', 'content' or 'numpy' is provided.
        """
        if self.path:
            return

        image_dir = settings.image_dir
        image_dir.mkdir(parents=True, exist_ok=True)
        file_path = image_dir / (self.media_id + ".bmp")

        if self.content:
            self._save_from_content(file_path)
        elif self.numpy is not None:
            self._save_from_numpy(file_path)
        elif self.url:
            self._save_from_url(file_path)
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        self.path = file_path
        self.is_saved = True

    def _save_from_numpy(self, file_path: Path):
        """Save the image from numpy on disk.

        Args:
            file_path (Path): The path of the file to write.
        """
        assert self.numpy is not None  # noqa: S101 TODO
        self.image_lib.write_file(file_path, self.numpy)

    def get_numpy(self) -> np.ndarray:
        """Load the image as a numpy array.

        Returns:
            np.ndarray: The image as a numpy array.

        Raises:
            ValueError: If none of 'path', 'url', 'content' or 'numpy' is provided.
            ImageReadingException: If there is an error reading the image.
        """
        if self.numpy is not None:
            return self.numpy
        elif self.path:
            self._load_numpy_from_path()
        elif self.url:
            self._load_numpy_from_url()
        elif self.content:
            self._load_numpy_from_content()
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        assert self.numpy is not None  # noqa: S101
        return self.numpy

    def get_pil_image(self) -> PIL.Image:
        """Get the image as a PIL image.

        Returns:
            PIL.Image: The image as a PIL image.
        """
        return PIL.Image.fromarray(self.get_numpy())

    def _load_numpy_from_path(self):
        """Load the image as a numpy array from a path.

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        assert self.path is not None  # noqa: S101
        try:
            self.numpy = self.image_lib.read_file(self.path)
        except Exception as e:
            raise ImageReadingException(self) from e

    def _load_numpy_from_image_bytes(self, img_bytes: bytes):
        """Load the image as a numpy array from image bytes (downloaded from URL or read from file).

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        try:
            self.numpy = self.image_lib.read_from_bytes(img_bytes)
        except Exception as e:
            raise ImageReadingException(self) from e

    def _load_numpy_from_url(self):
        """Load the image as a numpy array from a URL.

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        assert self.url is not None  # noqa: S101
        content: bytes = download_file(self.url)
        self._load_numpy_from_image_bytes(content)

    def _load_numpy_from_content(self):
        """Load the image as a numpy array from content.

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        assert self.content is not None  # noqa: S101
        self._load_numpy_from_image_bytes(self.content)

    def get_content(self) -> bytes:
        """Get the content of the image as bytes.

        Returns:
            bytes: The content of the image as bytes.

        Raises:
            ValueError: If none of 'path', 'url', 'content' or 'numpy' is provided.
        """
        if self.content:
            return self.content
        elif self.path:
            self._load_content_from_path()
        elif self.url:
            self._load_content_from_url()
        elif self.numpy is not None:
            self._load_content_from_numpy()
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        assert self.content is not None  # noqa: S101
        return self.content

    def _load_content_from_numpy(self):
        """Load the content of the image from numpy."""
        assert self.numpy is not None  # noqa: S101
        self.content = self.image_lib.write_to_bytes(self.numpy)

    def __repr__(self) -> str:
        """Get the representation of the image.

        Use md5 hash for the content of the image if it is available.

        For numpy array, use the shape of the array with the md5 hash of the array if it is available.

        Returns:
            str: The representation of the image.
        """
        content_hash = (
            hashlib.md5(self.content, usedforsecurity=False).hexdigest()
            if self.content
            else None
        )
        if self.numpy is not None:
            assert self.numpy is not None  # noqa: S101
            numpy_hash = hashlib.md5(
                self.numpy.tobytes(), usedforsecurity=False
            ).hexdigest()
            numpy_repr = f"ndarray(shape={self.numpy.shape}, dtype={self.numpy.dtype}, md5={numpy_hash})"
        else:
            numpy_repr = None
        return (
            f"Image(path={self.path}, "
            f"url={self.url}, "
            f"content={content_hash}, "
            f"numpy={numpy_repr}, "
            f"media_id={self.media_id})"
        )


class ImageInput(BaseModel):
    """An image input.

    Exactly one of 'path', 'url', or 'content' must be provided.

    If 'content' or 'numpy' is set to 'file',
    the image will be loaded from the files uploaded to the endpoint.


    Attributes:
        path (str): the file path of the image
        url (AnyUrl): the URL of the image
        content (bytes): the content of the image in bytes
        numpy (bytes): the image as a numpy array
    """

    path: str | None = Field(None, description="The file path of the image.")
    url: Annotated[
        AnyUrl | None,
        AfterValidator(lambda x: str(x) if x else None),
        Field(None, description="The URL of the image."),
    ]
    content: bytes | None = Field(
        None,
        description=(
            "The content of the image in bytes. "
            "Set this field to 'file' to upload files to the endpoint."
        ),
    )
    numpy: bytes | None = Field(
        None,
        description=(
            "The image as a numpy array. "
            "Set this field to 'file' to upload files to the endpoint."
        ),
    )
    media_id: MediaId = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The ID of the image. If not provided, it will be generated automatically.",
    )

    def set_file(self, file: bytes):
        """Sets the instance internal file data.

        If 'content' or 'numpy' is set to 'file',
        the image will be loaded from the file uploaded to the endpoint.

        set_file() should be called after the files are uploaded to the endpoint.

        Args:
            file (bytes): the file uploaded to the endpoint

        Raises:
            ValueError: if the content or numpy isn't set to 'file'
        """
        if self.content == b"file":
            self.content = file
        elif self.numpy == b"file":
            self.numpy = file
        else:
            raise ValueError(  # noqa: TRY003
                "The content or numpy of the image must be 'file' to set files."
            )

    def set_files(self, files: list[bytes]):
        """Set the files for the image.

        Args:
            files (List[bytes]): the files uploaded to the endpoint

        Raises:
            ValidationError: if the number of images and files aren't the same
        """
        if len(files) != 1:
            raise ValidationError.from_exception_data(
                title=self.__class__.__name__,
                line_errors=[
                    InitErrorDetails(
                        loc=("images",),
                        type="value_error",
                        ctx={
                            "error": ValueError(
                                "The number of images and files must be the same."
                            )
                        },
                        input=None,
                    )
                ],
            )
        self.set_file(files[0])

    @model_validator(mode="after")
    def check_only_one_field(self) -> Self:
        """Check that exactly one of 'path', 'url', 'content' or 'numpy' is provided.

        Raises:
            ValueError: if not exactly one of 'path', 'url', 'content' or 'numpy' is provided

        Returns:
            Self: the instance
        """
        count = sum(
            value is not None
            for value in [self.path, self.url, self.content, self.numpy]
        )
        if count != 1:
            raise ValueError(  # noqa: TRY003
                "Exactly one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        return self

    def convert_input_to_object(self) -> Image:
        """Convert the image input to an image object.

        Returns:
            Image: the image object corresponding to the image input

        Raises:
            ValueError: if the numpy file isn't set
        """
        if self.numpy and self.numpy != b"file":
            try:
                numpy = np.load(io.BytesIO(self.numpy), allow_pickle=False)
            except ValueError:
                raise ValueError("The numpy file isn't valid.")  # noqa: TRY003, TRY200, B904 TODO
        elif self.numpy == b"file":
            raise ValueError("The numpy file isn't set. Call set_files() to set it.")  # noqa: TRY003
        else:
            numpy = None

        return Image(
            path=Path(self.path) if self.path else None,
            url=self.url,
            content=self.content,
            numpy=numpy,
            media_id=self.media_id,
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "An image. \n"
                "Exactly one of 'path', 'url', or 'content' must be provided. \n"
                "If 'path' is provided, the image will be loaded from the path. \n"
                "If 'url' is provided, the image will be downloaded from the url. \n"
                "The 'content' will be loaded automatically "
                "if files are uploaded to the endpoint (should be set to 'file' for that)."
            )
        },
        validate_assignment=True,
        file_upload=True,
        file_upload_description="Upload image file.",
    )


class ImageInputList(BaseListModel):
    """A pydantic model for a list of images to be used as input.

    Only used for the requests, DO NOT use it for anything else.
    Convert it to a list of image objects with convert_input_to_object().
    """

    root: list[ImageInput]

    @model_validator(mode="after")
    def check_non_empty(self) -> Self:
        """Check that the list of images isn't empty.

        Raises:
            ValueError: if the list of images is empty

        Returns:
            Self: the instance
        """
        if len(self.root) == 0:
            raise ValueError("The list of images must not be empty.")  # noqa: TRY003
        return self

    def set_files(self, files: list[bytes]):
        """Set the files for the images.

        Args:
            files (List[bytes]): the files uploaded to the endpoint

        Raises:
            ValidationError: if the number of images and files aren't the same
        """
        if len(self.root) != len(files):
            error = ValueError("The number of images and files must be the same.")
            # raise ValidationError(error,
            raise error
        for image, file in zip(self.root, files, strict=False):
            image.set_file(file)

    def convert_input_to_object(self) -> list[Image]:
        """Convert the list of image inputs to a list of image objects.

        Returns:
            List[Image]: the list of image objects corresponding to the image inputs
        """
        return [image.convert_input_to_object() for image in self.root]

    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "A list of images. \n"
                "Exactly one of 'path', 'url', or 'content' must be provided for each image. \n"
                "If 'path' is provided, the image will be loaded from the path. \n"
                "If 'url' is provided, the image will be downloaded from the url. \n"
                "The 'content' will be loaded automatically "
                "if files are uploaded to the endpoint (should be set to 'file' for that)."
            )
        },
        file_upload=True,
        file_upload_description="Upload image files.",
    )
